const inflection = require('inflection');
const stringUtils = require('ember-cli-string-utils');
const EOL = require('os').EOL;
const path = require('path');
const fs = require('fs');
const nbFakeData = 6;

module.exports = {
  description: 'Generates an ember-data model and the associated Mirage couterpart and screens.',

  anonymousOptions: ['name', 'attr:type'],

  fileMapTokens(options) {
    if (options.pod) {
      return {
        // Never tested. Not used to use pod structure ! 
        __root__() {
          return 'app';
        },
        __path__() {
          return options.blueprintName;
        },
        __name__() {
          return options.dasherizedModuleName;
        },
        __mirage__() {
          return '../mirage'
        },
        __cloud__() {
          return '../cloud'
        }
      }
    } else {
      return {
        __root__() {
          if (options.inDummy) {
            return 'tests/dummy/app';
          } else {
            return 'app';
          }
        },
        __path__() {
          return 'models';
        },
        __name__() {
          return options.dasherizedModuleName;
        },
        __mirage__() {
          return '../mirage'
        },
        __cloud__() {
          return '../cloud'
        }
      }
    }
  },

  locals(options) {
    let camelizedName = stringUtils.camelize(options.entity.name);

    /*
        for (let name in entityOptions) {
          let type = entityOptions[name] || '';
          let foreignModel = name;
          let foreignKey = 'id';
          if (type.indexOf(':') > -1) {
            foreignModel = type.split(':')[1];
            foreignKey = type.split(':')[2];
            type = type.split(':')[0];
          }
          let dasherizedName = stringUtils.dasherize(name);
          let camelizedName = stringUtils.camelize(name);
          let dasherizedType = stringUtils.dasherize(type);
          let dasherizedForeignModel = stringUtils.dasherize(foreignModel);
          let dasherizedForeignModelSingular = inflection.singularize(dasherizedForeignModel);
    */
    return {
      singularEntityName: camelizedName,
      entityValidation: entityValidation(camelizedName, options.entity.options),
      entityModel: entityModel(camelizedName, options.entity.options),
      mirageFactory: mirageFactory(camelizedName, options.entity.options),
      mirageModel: mirageModel(camelizedName, options.entity.options),
      controllerInitEntity: controllerInitEntity(camelizedName, options.entity.options),
      controllerCreateEntity: controllerCreateEntity(camelizedName, options.entity.options),
      controllerModelTable: controllerModelTable(camelizedName, options.entity.options),
      templateEntityForm: templateEntityForm(camelizedName, options.entity.options)
    };
  },

  afterInstall(options) {
    let blueprintName = options.originBlueprintName;
    let entityName = options.entity.name;

    // Check that /mirage/serializers/applications.js is existing
    let mirageSerializerPath = (options.dummy) ? "tests/dummy/mirage/serializers/application.js" : "mirage/serializers/application.js";
    if (!fs.existsSync(mirageSerializerPath)) {
      this.ui.writeLine("Creating file " + mirageSerializerPath);
      let serializerContent = "import { JSONAPISerializer } from 'ember-cli-mirage';\n\n\texport default JSONAPISerializer.extend({\n});\n";
      fs.writeFileSync(mirageSerializerPath, serializerContent, 'utf-8', 'w+');
    }

    // Complete /mirage/scenarios/default.js
    let mirageScenarioPath = (options.dummy) ? "tests/dummy/mirage/scenarios/default.js" : "mirage/scenarios/default.js";
    let scenarioLine = '\tserver.createList\(\'' + entityName + '\', ' + nbFakeData + '\);\n';
    if (!fs.existsSync(mirageScenarioPath)) {
      this.ui.writeLine("Creating file " + mirageScenarioPath);
      let scenarioContent = "export default function(server) {\n" + scenarioLine + "}\n";
      fs.writeFileSync(mirageScenarioPath, scenarioContent, 'utf-8', 'w+');
    } else {
      addLineToFile(this, mirageScenarioPath, /export default function\(server\) {/, scenarioLine);
    }

    // Complete /mirage/config.js
    let mirageConfigPath = (options.dummy) ? "tests/dummy/mirage/config.js" : "mirage/config.js";
    let configLine = "\t\tthis.get('/" + inflection.pluralize(entityName) + "');\n" +
      "\t\tthis.get('/" + inflection.pluralize(entityName) + "/:id');\n" +
      "\t\tthis.delete('/" + inflection.pluralize(entityName) + "/:id');\n" +
      "\t\tthis.patch('/" + inflection.pluralize(entityName) + "/:id');\n" +
      "\t\tthis.post('/" + inflection.pluralize(entityName) + "');";
    if (!fs.existsSync(mirageConfigPath)) {
      this.ui.writeLine("Creating file " + mirageConfigPath);
      let configContent = "import ENV from './../config/environment';\n" +
        "export default function() {\n" +
        "\tthis.namespace = '';\n" +
        "\tif (!ENV.APP.proxy) {\n" + configLine + "\n\t} else {\n" +
        "\t\tthis.passthrough();\n\t}\n}\n";
      fs.writeFileSync(mirageConfigPath, configContent, 'utf-8', 'w+');
    } else {
      addLineToFile(this, mirageConfigPath, /if \(!ENV\.APP\.proxy\) {/, configLine);
    }

    // Add route in router.js for the entity page
    let routerPath = (options.dummy) ? "tests/dummy/app/router.js" : "app/router.js";
    let routeName = '\tthis.route\(\'' + entityName + '\'\);';
    addLineToFile(this, routerPath, /this\.route\('entity-factory', function\(\) {/, routeName);

    // Add Entity select entry in the entity controller file
    let entityIndexPath = (options.dummy) ? "tests/dummy/app/controllers/entity-factory.js" : "app/controllers/entity-factory.js";
    let entityIndexEntry = "'" + entityName + "',";
    addLineToFile(this, entityIndexPath, /entities: \[/, entityIndexEntry);
  }
};

function addLineToFile(ctx, filePath, markerString, addedLine) {
  let fileContents = fs.readFileSync(filePath, 'utf-8');

  if (fileContents.indexOf(addedLine) === -1) {
    ctx.ui.writeLine("\tAdd line " + addedLine + " to file " + filePath);
    fileContents = fileContents.replace(markerString, [
      '$&',
      addedLine,
    ].join('\n'));
  }
  if (fileContents.indexOf(addedLine) === -1) {
    ctx.ui.writeWarnLine(
      'Unable to update ' + filePath + '. You should update this file manually.');
  } else {
    fs.writeFileSync(filePath, fileContents, 'utf-8');
  }
};

function templateEntityForm(name, options) {
  let form = ['{{#bs-form formLayout="vertical" model=newEntry onSubmit=(action "submit") as |form|}}'];
  let attributes = [];
  for (var prop in options) {
    let type = options[prop].split(':')[0];
    let targetEntity = options[prop].split(':')[1];
    switch (type) {
      case "string":
      case "number":
        form.push('{{form.element controlType="text" label=\"' + prop + '\" property=\"'+ prop + '\" value=' +
                 prop + ' onChange=(action (mut ' + prop + ')) placeholder=\"' + prop + '\"}}');
        break;

      case "boolean":
        form.push('{{form.element controlType="checkbox" label=\"' + prop + '\" property=\"'+ prop + '\" value=' + 
                prop + ' onChange=(action (mut ' + prop + ')) placeholder=\"' + prop + '\"}}');
        break;

      case "date":
        form.push('<div class="row form-group"><label class="col-form-label col-md-4">' + prop +
          '</label>{{pikaday-input useUTC=true format="DD/MM/YYYY" onSelection=(action (mut ' + prop + '))}}</div>');
        break;

      case 'belongs-to':
        // todo How to deal with relationships
        break;

      case 'has-many':
        // todo How to deal with relationships
        break;

      default:
        // do nothing
        break;
      }
    }
    form.push("{{/bs-form}}");
    return form.join('\n');
  };

function templateEntityForm2(name, options) {
  let form = ['{{#bs-form formLayout="vertical" model=this onSubmit=(action "submit") as |form|}}'];
  let attributes = [];
  for (var prop in options) {
    let type = options[prop].split(':')[0];
    let targetEntity = options[prop].split(':')[1];
    switch (type) {
      case "string":
      case "number":
        form.push('{{form.element controlType="text" label=\"' + prop + '\" value=' + prop + ' onChange=(action (mut ' + prop + ')) placeholder=\"' + prop + '\"}}');
        break;

      case "boolean":
        form.push('{{form.element controlType="checkbox" label=\"' + prop + '\" value=' + prop + ' onChange=(action (mut ' + prop + ')) placeholder=\"' + prop + '\"}}');
        break;

      case "date":
        form.push('<div class="row form-group"><label class="col-form-label col-md-4">' + prop +
          '</label>{{pikaday-input useUTC=true format="DD/MM/YYYY" onSelection=(action (mut ' + prop + '))}}</div>');
        break;

      case 'belongs-to':
        // todo How to deal with relationships
        break;

      case 'has-many':
        // todo How to deal with relationships
        break;

      default:
        // do nothing
        break;
      }
    }
    form.push("{{/bs-form}}");
    return form.join('\n');
  };

  function mirageFactory(name, options) {
    let content = "import {\n\tFactory,\n\tfaker\n\t} from 'ember-cli-mirage';\n\nexport default Factory.extend({\n";
    let attributes = [];
    for (var prop in options) {
      let type = options[prop].split(':')[0];
      let constraintsString = options[prop].split(':')[1];
      switch (type) {
        case "string":
          attributes.push("\t" + prop + "() {\n\t\treturn faker.name.findName();\n\t}");
          break;

        case "date":
          attributes.push("\t" + prop + "() {\n\t\treturn faker.date.past();\n\t}");
          break;

        case "number":
          let constraints = [];
          let min = 0,
            max = 10000;
          if (constraintsString) {
            constraints = constraintsString.split(',');
            for (var i = 0; i < constraints.length; i++) {
              let constraint = constraints[i];
              if (constraint.startsWith("min")) {
                min = constraint.substring(constraint.indexOf('(') + 1, constraint.indexOf(')'));
              } else if (constraint.startsWith("max")) {
                max = constraint.substring(constraint.indexOf('(') + 1, constraint.indexOf(')'));
              }
            }
          }
          attributes.push("\t" + prop + "() {\n\t\treturn faker.random.number({min: " + min + ",max: " + max + "});\n\t}");
          break;

        case "boolean":
          attributes.push("\t" + prop + "() {\n\t\treturn faker.random.boolean();\n\t}");
          break;

        default:
          // do nothing
          break;
      }
    }
    content += attributes.join(',\n') + "\n});";

    return content;
  };

  function mirageModel(name, options) {
    let content = "import { Model, belongsTo, hasMany} from 'ember-cli-mirage';\n\nexport default Model.extend({\n";
    let attributes = [];
    for (var prop in options) {
      let type = options[prop].split(':')[0];
      switch (type) {
        case 'belongs-to':
          attributes.push("\t" + prop + ": belongsTo()");
          break;

        case 'has-many':
          attributes.push("\t" + prop + ": hasMany()");
          break;

        default:
          // do nothing
          break;
      }
    }
    content += attributes.join(',\n') + "\n});";

    return content;
  };
  
  function controllerInitEntity(name, options) {
    let attributes = [];
    for (var prop in options) {
      let type = options[prop].split(':')[0];
      switch (type) {
        case "string":
          attributes.push(prop + ": ''");
          break;

        case "number":
          attributes.push(prop + ": 0");
          break;
  
        case "boolean":
          attributes.push(prop + ": 'false'");
          break;
  
        case "date":
          attributes.push(prop + ": new Date()");
          break;
  
        case 'belongs-to':
          attributes.push(prop + ": ''");
          break;
  
        case 'has-many':
          attributes.push(prop + ": []");
          break;
  
        default:
          // do nothing
          break;
        }
    }
    return attributes.join(',');
  };

  function controllerCreateEntity(name, options) {
    let attributes = [];
    for (var prop in options) {
      if (prop)
        attributes.push(prop + ": this.newEntry." + prop);
    }
    return attributes.join(',');
  };

  function controllerModelTable(name, options) {
    let content = "";
    let tablecols = [];
    let attributes = [];
    for (var prop in options) {
      let type = options[prop].split(':')[0];
      let targetEntity = options[prop].split(':')[1];
      switch (type) {
        case "string":
        case "date":
        case "number":
        case "boolean":
          tablecols.push('\t\tEmberObject.create({\n\t\t\tpropertyName: \"' + prop + '\",\n\t\t\ttitle: \"' + prop + '\"\n\t\t})');
          attributes.push("\t" + prop + ": null,");
          break;

        case 'belongs-to':
          // todo How to deal with relationships
          attributes.push("\t" + prop + ": null,");
          break;

        case 'has-many':
          // todo How to deal with relationships
          attributes.push("\t" + prop + ": null,");
          break;

        default:
          // do nothing
          break;
      }
    }
    content += attributes.join('\n') + name + "TableColumns: computed(function() {\n\tvar col = A([\n";
    content += tablecols.join(',\n') + "]);\n\tcol.pushObject({\n\t\ttitle: 'Delete',\n\t\tcomponent: 'delete-row'\n\t});\n\treturn col;\t}),\n";
    content += name + "TableContent: computed(function() {\n\treturn this.get('model');\n}),";

    return content;
  };

  function entityModel(name, options) {
    let content = "import DS from 'ember-data';\n\nexport default DS.Model.extend({\n";
    let attributes = [];
    for (var prop in options) {
      let type = options[prop].split(':')[0];
      let targetEntity = options[prop].split(':')[1];
      switch (type) {
        case "string":
        case "date":
        case "number":
        case "boolean":
          attributes.push("\t" + prop + ": DS.attr('" + type + "')");
          break;

        case 'belongs-to':
          attributes.push("\t" + prop + ": DS.belongsTo('" + targetEntity + "')");
          break;

        case 'has-many':
          attributes.push("\t" + prop + ": DS.hasMany('" + targetEntity + "')");
          break;

        default:
          // do nothing
          break;
      }
    }
    content += attributes.join(',\n') + "\n});";

    return content;
  };

  function entityValidation(name, options) {
    let content = "import {validatePresence, validateLength} from 'ember-changeset-validations/validators';\n\n" +
      "export default {\n";
    for (var prop in options) {
      let attr = options[prop].split(':')[1];
      if (attr) { // there is at least one constraint
        let constraints = [];
        let blocks = attr.split(',');
        for (var i = 0; i < blocks.length; i++) {
          let constraint = blocks[i];
          if (constraint.startsWith('required')) {
            constraints.push('validatePresence(true)');
          }
        }
        if (constraints.length>0)
          content += "\t" + prop + ": [" + constraints.join(',') + "],\n"
      }
    }
    content += "}\n"
    return content;
  }

  function dsAttr(name, type) {
    switch (type) {
      case 'belongs-to':
        return "DS.belongsTo('" + name + "')";
      case 'has-many':
        return "DS.hasMany('" + name + "')";
      case '':
        //"If you don't specify the type of the attribute, it will be whatever was provided by the server"
        //https://emberjs.com/guides/models/defining-models/
        return 'DS.attr()';
      default:
        return "DS.attr('" + type + "')";
    }
  };

  const fakerize = (type) => {
    let typeFake
    switch (type) {
      case 'boolean':
        typeFake = 'faker.random.boolean()'
        break
      case 'date':
        typeFake = 'faker.date.past()'
        break
      case 'number':
        typeFake = 'faker.random.number({min: 0,max: 10000})'
        break
      case 'string':
        typeFake = 'faker.name.findName()'
        break
      default:
        return null
    }
    return ' () {\n\t\treturn ' + typeFake + '\n\t}'
  };

  const relFunction = (str) => {
    if (/has-many/.test(str)) {
      return ': hasMany()'
    }
    if (/belongs-to/.test(str)) {
      return ': belongsTo()'
    }
  }